package es.prodevelop.cit.gvsig.arcims.fmap.layers;

import java.io.File;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.cresques.cts.ICoordTrans;
import org.cresques.cts.IProjection;
import org.cresques.cts.ProjectionPool;
import org.cresques.cts.gt2.CoordSys;
import org.cresques.cts.gt2.CoordTrans;

import com.hardcode.driverManager.Driver;
import com.hardcode.driverManager.DriverLoadException;
import com.iver.cit.gvsig.fmap.FMap;
import com.iver.cit.gvsig.fmap.drivers.DefaultDBDriver;
import com.iver.cit.gvsig.fmap.drivers.DriverIOException;
import com.iver.cit.gvsig.fmap.drivers.VectorialDatabaseDriver;
import com.iver.cit.gvsig.fmap.drivers.VectorialDriver;
import com.iver.cit.gvsig.fmap.drivers.VectorialFileDriver;
import com.iver.cit.gvsig.fmap.layers.CancelationException;
import com.iver.cit.gvsig.fmap.layers.FLayer;
import com.iver.cit.gvsig.fmap.layers.FLayers;
import com.iver.cit.gvsig.fmap.layers.FLyrRaster;
import com.iver.cit.gvsig.fmap.layers.FLyrVect;
import com.iver.cit.gvsig.fmap.layers.LayerFactory;
import com.iver.cit.gvsig.fmap.layers.SelectionEvent;
import com.iver.cit.gvsig.fmap.layers.SelectionListener;
import com.iver.cit.gvsig.fmap.layers.XMLException;
import com.iver.cit.gvsig.fmap.layers.layerOperations.Selectable;
import com.iver.cit.gvsig.fmap.rendering.VectorialLegend;
import com.iver.utiles.IPersistance;
import com.iver.utiles.XMLEntity;

import es.prodevelop.cit.gvsig.arcims.gui.wizards.ArcImsWizard;

public class FFeatureLyrArcIMSCollection extends FLayers implements SelectionListener {
	
	private static Logger logger = Logger.getLogger(FFeatureLyrArcIMSCollection.class.getName());
	private ArrayList layerlist = new ArrayList();
	
	public FFeatureLyrArcIMSCollection(FMap fmap, FLayers parent) {
		super(fmap, parent);
	}
	
	public FFeatureLyrArcIMSCollection() {
		super(null, null);
	}
	
	public FMap getFMap() {
		if (getParentLayer() != null) {
			return getParentLayer().getFMap();
		} else {
			return null;
		}
	}
	
	public void addLayer(FLayer layer) throws CancelationException {
		if (! (layer instanceof FFeatureLyrArcIMS)) {
			logger.error("Only FFeatureLyrArcIMS layers allowed in this collection. ");
			return;
		}
		FFeatureLyrArcIMS lyr = (FFeatureLyrArcIMS) layer;
		lyr.addSelectionListener(this);
		super.addLayer(layer);
		layerlist.add(layer);
	}

	public void selectionChanged(SelectionEvent e) {
		// e.get
		getFMap().invalidate();
		// System.out.println("==========> Selection changed somewhere....");
		// TODO Auto-generated method stub
		
	}
	
	public void setXMLEntity(XMLEntity xml) throws XMLException{
		
		setActive(xml.getBooleanProperty("active"));
		setName(xml.getStringProperty("name"));
		setMinScale(xml.getDoubleProperty("minScale"));
		setMaxScale(xml.getDoubleProperty("maxScale"));
		setVisible(xml.getBooleanProperty("visible"));
		
		if (xml.contains("proj")) {
			setProjection(ProjectionPool.get(xml.getStringProperty("proj")));
		}
		
		if (xml.contains("transparency")) {
			setTransparency(xml.getIntProperty("transparency"));
		}

		// ------------
		
		int numLayers = xml.getIntProperty("numLayers");
		String[] s = xml.getStringArrayProperty("LayerNames");
		
		for (int i = 0; i < numLayers; i++) {
			FLayer layer = null;
			try {
				String className = xml.getChild(i).getStringProperty("className");
				Class clase = Class.forName(className);
				layer = (FLayer) clase.newInstance();
				layer.setName(s[i]);
				layer.setXMLEntity(xml.getChild(i));
				layer.load();
				logger.debug("Layer: "+ layer.getName() +" has been loaded.");
			} catch (Exception e) {
				logger.error("While loading layer: " + layer.getName() + " (visible = false)", e);
				layer.setVisible(false);
			}
			addLayer(layer);
		}
	}
	
	public XMLEntity getXMLEntity() throws XMLException {
		
		XMLEntity xml = new XMLEntity();
		xml.putProperty("className", this.getClass().getName());
		xml.putProperty("active", isActive());
		xml.putProperty("name", getName());
		xml.putProperty("minScale", getMinScale());
		xml.putProperty("maxScale", getMaxScale());
		xml.putProperty("visible", isVisible());
		
		if (getProjection() != null) {
			xml.putProperty("proj", getProjection().getAbrev());
		}
		xml.putProperty("transparency", getTransparency());
		
		// --------------------

		xml.putProperty("numLayers", layerlist.size());

		String[] s = new String[layerlist.size()];

		for (int i = 0; i < layerlist.size(); i++) {
			s[i] = ((FFeatureLyrArcIMS) layerlist.get(i)).getName();
		}

		xml.putProperty("LayerNames", s);

		for (int i = 0; i < layerlist.size(); i++) {
			xml.addChild(((FFeatureLyrArcIMS) layerlist.get(i)).getXMLEntity());
		}

		return xml;
	}
	
	
	
//	public boolean isActive() {
//		for (int i=0; i<getLayersCount(); i++) {
//			FLayer lyr = getLayer(i);
//			if (lyr.isActive()) return true;
//		}
//		return false;
//	}
	
	
	
}
	
	
//	private ArrayList layers;
//	// array of FFeatureLyrArcIMSCollectionItem objects
//	// FFeatureLyrArcIMSCollectionItem: the id will always be the index in
//	// the arraylist:
//	// layers.get(i).getId() == i
//	
//	public FFeatureLyrArcIMSCollection() {
//		layers = new ArrayList();
//	}
//
//	public Rectangle2D getFullExtent() throws DriverException {
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FFeatureLyrArcIMS layer;
//		Rectangle2D acumExtent = null;
//
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			if (acumExtent == null) {
//				acumExtent = layer.getFullExtent();
//			} else {
//				acumExtent = acumExtent.createUnion(layer.getFullExtent());
//			}
//		}
//		if (acumExtent == null) {
//			DriverException de = new DriverException("Unable to get Extent from FFeatureLyrArcIMSCollection.");
//			throw de;
//		}
//		return acumExtent;
//	}
//
//	public void draw(BufferedImage image, Graphics2D g, ViewPort viewPort,
//			Cancellable cancel, double scale) throws DriverException {
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FFeatureLyrArcIMS layer;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			layer.draw(image, g, viewPort, cancel, scale);
//		}
//	}
//
//	public void print(Graphics2D g, ViewPort viewPort, Cancellable cancel,
//			double scale) throws DriverException {
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FFeatureLyrArcIMS layer;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			layer.print(g, viewPort, cancel, scale);
//		}
//	}
//
//	public void addLayerCollectionListener(LayerCollectionListener listener) {
//		System.out.println("============================================================================");
//		System.out.println("public void addLayerCollectionListener(LayerCollectionListener listener):");
//		System.out.println("listener = " + listener.toString());
//		System.out.println("============================================================================");
//	}
//
//	public void removeLayerCollectionListener(LayerCollectionListener listener) {
//		// TODO Auto-generated method stub
//
//	}
//
//	public void addLayer(FLayer layer) throws CancelationException {
//		if (! (layer instanceof FFeatureLyrArcIMS)) {
//			System.err.println("Only FFeatureLyrArcIMS layers allowed in this collection");
//			return;
//		}
//		String theName = layer.getName();
//		int theId = layers.size();
//		FFeatureLyrArcIMS theLayer = (FFeatureLyrArcIMS) layer;
//		FFeatureLyrArcIMSCollectionItem item =
//			new FFeatureLyrArcIMSCollectionItem(theLayer, theId, theName);
//		layers.add(item);
//		updateIds();
//	}
//
//	public void moveTo(int from, int to) throws CancelationException {
//		FFeatureLyrArcIMSCollectionItem item =
//			(FFeatureLyrArcIMSCollectionItem) layers.remove(from);
//		layers.add(to, item);
//		updateIds();
//	}
//
//	public void removeLayer(FLayer lyr) throws CancelationException {
//		FLayer layer;
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			if (layer == lyr) {
//				removeLayer(i);
//				break;
//			}
//		}
//		updateIds();
//	}
//
//	public void removeLayer(int idLayer) {
//		// the id will always be the index in the arraylist:
//		// layers.get(i).getId() == i
//		layers.remove(idLayer);
//		updateIds();
//	}
//
//	public void removeLayer(String layerName) {
//		String name;
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			name = item.getName();
//			if (name.compareToIgnoreCase(layerName) == 0) {
//				removeLayer(i);
//				break;
//			}
//		}
//		// updateIds();
//	}
//
//	public FLayer[] getVisibles() {
//		ArrayList resultArrayList = new ArrayList();
//		FFeatureLyrArcIMS layer;
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			if (layer.isVisible()) resultArrayList.add(layer);
//		}
//		
//		int size = resultArrayList.size();
//		FLayer result[] = new FLayer[size];
//		for (int i=0; i<size; i++) {
//			result[i] = (FLayer) resultArrayList.get(i);
//		}
//		return result;
//	}
//
//	public FLayer[] getActives() {
//		ArrayList resultArrayList = new ArrayList();
//		FFeatureLyrArcIMS layer;
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			if (layer.isActive()) resultArrayList.add(layer);
//		}
//		
//		int size = resultArrayList.size();
//		FLayer result[] = new FLayer[size];
//		for (int i=0; i<size; i++) {
//			result[i] = (FLayer) resultArrayList.get(i);
//		}
//		return result;
//	}
//
//	public FLayer getLayer(int index) {
//		FFeatureLyrArcIMSCollectionItem item;
//		FLayer layer;
//		item = (FFeatureLyrArcIMSCollectionItem) layers.get(index);
//		layer = item.getLayer();
//		return layer; 
//	}
//
//	public FLayer getLayer(String layerName) {
//		String name;
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FLayer layer;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			name = item.getName();
//			if (name.compareToIgnoreCase(layerName) == 0) {
//				layer = (FLayer) item.getLayer();
//				return layer;
//			}
//		}
//		System.err.println("Unable to find FLayer by name: " + layerName);
//		return null;
//	}
//
//	public int getLayersCount() {
//		return layers.size();
//	}
//
//	public void setAllActives(boolean active) {
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FFeatureLyrArcIMS layer;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			layer.setActive(true);
//		}
//	}
//
//	public void setAllVisibles(boolean visible) {
//		int count = layers.size();
//		FFeatureLyrArcIMSCollectionItem item;
//		FFeatureLyrArcIMS layer;
//		
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			layer = item.getLayer();
//			layer.setVisible(true);
//		}
//	}
//	
//	private void updateIds() {
//		FFeatureLyrArcIMSCollectionItem item;
//		int count = layers.size();
//		for (int i=0; i<count; i++) {
//			item = (FFeatureLyrArcIMSCollectionItem) layers.get(i);
//			item.setId(i);
//		}
//	}