/**
 * 
 */
package es.prodevelop.cit.gvsig.arcims.gui.toc;

import java.awt.event.ActionEvent;
import java.util.ArrayList;

import javax.swing.JMenuItem;

import org.cresques.io.GeoRasterFile;
import org.cresques.io.raster.RasterFilterStackManager;
import org.cresques.px.PxRaster;
import org.cresques.ui.raster.BandSetupPanel;
import org.cresques.ui.raster.EnhancedPanel;
import org.cresques.ui.raster.FilterRasterDialogPanel;
import org.cresques.ui.raster.RasterTransparencyPanel;

import com.iver.andami.PluginServices;
import com.iver.cit.gvsig.fmap.layers.FLayer;

import es.prodevelop.cit.gvsig.arcims.fmap.layers.FRasterLyrArcIMS;
import es.prodevelop.cit.gvsig.arcims.gui.dialogs.ArcImsRasterPropsDialog;

import com.iver.cit.gvsig.gui.toc.FPopupMenu;
import com.iver.cit.gvsig.gui.toc.TocMenuEntry;

/**
 * This class implements the raster properties menu item that is
 * added to the ArcIMS layer's popup menu.
 *   
 * @author jldominguez
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public class ArcImsRasterPropsTocMenuEntry extends TocMenuEntry {
	private JMenuItem propsMenuItem;
	FLayer lyr = null;
	private ArcImsRasterPropsDialog	propsDialog = null;
	private BandSetupPanel bandSetup = null;

	public void initialize(FPopupMenu m) {
		super.initialize(m);
		
		if (isTocItemBranch()) {
			lyr = getNodeLayer();
    		// ArcIMS layer
    		if ((lyr instanceof FRasterLyrArcIMS)) {
    			propsMenuItem = new JMenuItem(PluginServices.getText(this, "propiedades_raster"));
    			getMenu().add( propsMenuItem );
    			propsMenuItem.setFont(FPopupMenu.theFont);
    			getMenu().setEnabled(true);
    			propsMenuItem.addActionListener(this);   			
     		}
		}
	}

	/**
	 * Creates an ArcImsRasterPropsDialog object and adds it to the MDIManager.
	 */
	public void actionPerformed(ActionEvent e) {
		lyr = getNodeLayer();
        
		if(lyr instanceof FRasterLyrArcIMS && ((FRasterLyrArcIMS) lyr).getPxRaster()!=null){
			RasterFilterStackManager stackManager = null;
			
			FRasterLyrArcIMS layer = (FRasterLyrArcIMS) lyr;

			stackManager = new RasterFilterStackManager(layer.getFilterStack());
			
			int[][] rangeR = stackManager.getTransparecyR();
			int[][] rangeG = stackManager.getTransparecyG();
			int[][] rangeB = stackManager.getTransparecyB();
			
			propsDialog = new ArcImsRasterPropsDialog(layer, rangeR, rangeG, rangeB);
		       			
			int alpha = layer.getPxRaster().getAlpha();
			
			bandSetup = ((FilterRasterDialogPanel)propsDialog.getContentPane()).getBandSetup();
			GeoRasterFile[] files = layer.getPxRaster().getFiles();
			bandSetup.addFiles(files);
			for(int i=0; i< files.length;i++) 
				propsDialog.addNumBands(files[i].getBandCount());
				
			PxRaster px = layer.getPxRaster();
			int posR = 0, posG = 0, posB = 0;
			
			for(int i=0;i<px.getPosFile(GeoRasterFile.RED_BAND);i++)
				posR += files[i].getBandCount();
			posR += px.getBand(GeoRasterFile.RED_BAND);
			
			for(int i=0;i<px.getPosFile(GeoRasterFile.GREEN_BAND);i++)
				posG += files[i].getBandCount();
			posG += px.getBand(GeoRasterFile.GREEN_BAND);
			
			for(int i=0;i<px.getPosFile(GeoRasterFile.BLUE_BAND);i++)
				posB += files[i].getBandCount();
			posB += px.getBand(GeoRasterFile.BLUE_BAND);
			
					
			bandSetup.assignBand(posR, GeoRasterFile.RED_BAND);
			bandSetup.assignBand(posG, GeoRasterFile.GREEN_BAND);
			bandSetup.assignBand(posB, GeoRasterFile.BLUE_BAND);	
									
			// Sets current alpha value to the dialog
			RasterTransparencyPanel rasterTrans = ((FilterRasterDialogPanel)propsDialog.getContentPane()).getTransparencyPanel();
			rasterTrans.setOpacity(alpha);
				
			// Sets current enhacement to the dialog
			EnhancedPanel pEnhan =  ((FilterRasterDialogPanel)propsDialog.getContentPane()).getEnhancedPanel();
			
			propsDialog.setRasterFilterStackManager(stackManager);
			int tail = 0;
			if(	pEnhan != null &&
				stackManager.getStackStats() != null){
				tail = (int)(stackManager.getStackStats().tailPercent*200);
				pEnhan.setPercentTail(tail);
			}
			if(stackManager.isActive(stackManager.getTypeFilter("enhanced")))
				pEnhan.setSelectedFilter(1);
			else
				pEnhan.setSelectedFilter(0);
			
			bandSelector(stackManager, posR, posG, posB);
			
			propsDialog.readStat();
			
			PluginServices.getMDIManager().addView(propsDialog);       	
		}
	}
	
	/**
	 * Selects the active bands in the band panel
	 * from the filters retrieved from the FilterStack
	 *  
	 * @param stackManager RasterFilterStackManager
	 * @param posR R-band position
	 * @param posG G-band position
	 * @param posB B-band position
	 */
	private void bandSelector(RasterFilterStackManager stackManager, int posR, int posG, int posB){
		ArrayList stackList = stackManager.getStringsFromStack();
		String hideBands = null;
		for(int i=0;i<stackList.size();i++){
			if(((String)stackList.get(i)).startsWith("filter.removebands.bands"))
				hideBands = stackManager.getValue((String)stackList.get(i));
		}
		
		// 1, 2 or 3 bands to show
		if(hideBands != null){
			int pos = 2;
			if(hideBands.length() == 1)
				pos = 1;
			else if(hideBands.length() == 2)
				pos = 0;
			bandSetup.getFileList().getJComboBox().setSelectedIndex(pos);
			
			// Reset table controls
			
			for(int i=0;i<bandSetup.getRGBTable().getModel().getRowCount();i++)
				for(int j=0;j<3;j++)
					bandSetup.getRGBTable().getModel().setValueAt(new Boolean(false), i, j);
			
			if(hideBands.equals("GB") || hideBands.equals("G") || hideBands.equals("B") )
				bandSetup.getRGBTable().getModel().setValueAt(new Boolean(true), posR, 0);
			
			if(hideBands.equals("RB") || hideBands.equals("R") || hideBands.equals("B") )
				bandSetup.getRGBTable().getModel().setValueAt(new Boolean(true), posG, 1);
			
			if(hideBands.equals("RG") || hideBands.equals("R") || hideBands.equals("G") )
				bandSetup.getRGBTable().getModel().setValueAt(new Boolean(true), posB, 2);
		}
	}
	

}
