/**
 * 
 */
package org.gvsig.remoteClient.arcims;

import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.gvsig.remoteClient.arcims.exceptions.ArcImsException;
import org.gvsig.remoteClient.arcims.utils.ArcImsDownloadUtils;
import org.gvsig.remoteClient.arcims.utils.ArcImsValueFactory;
import org.gvsig.remoteClient.arcims.utils.FieldInformation;
import org.gvsig.remoteClient.arcims.utils.GetFeaturesTags;
import org.gvsig.remoteClient.arcims.utils.GetVariables;
import org.gvsig.remoteClient.arcims.utils.ServiceInfoTags;
import org.gvsig.remoteClient.arcims.utils.ServiceInformation;
import org.gvsig.remoteClient.arcims.utils.ServiceInformationLayerFeatures;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

import com.hardcode.gdbms.engine.values.Value;
import com.iver.cit.gvsig.fmap.core.FPolygon2D;
import com.iver.cit.gvsig.fmap.core.GeneralPathX;
import com.iver.cit.gvsig.fmap.core.IGeometry;
import com.iver.cit.gvsig.fmap.core.ShapeFactory;

/**
 * @author jsanz
 *
 */
public class ArcImsCache {
	private static Logger logger = Logger.getLogger(ArcImsCache.class.getName());
//	private final int MAX_STATUS_WIDTH = 1200;
	private final int MIN_STATUS_WIDTH = 200;
	private int num_accuracies;
	
	
	/**
	 * Double array of accuracies to be used in the cache
	 */
	private double[] accuracies;
	/**
	 * Full extent of the layer cached
	 */
	private Rectangle2D fullextent;
	/**
	 * FPolygon2D array of visited windows, one for every 
	 */
	private FPolygon2D[] visitedWindows;
	
	private HashMap[] features;
	
	/**
	 * To create a Cache, we need an maximum extent and the number of accuracies to take account 
	 * @param extent
	 * @param num_accuracies
	 */

	public ArcImsCache(Rectangle2D extent, int num_accuracies){
		this.fullextent = extent;
		this.num_accuracies = num_accuracies;
		this.accuracies = new double[num_accuracies];
		this.features = new HashMap[num_accuracies];
		this.createAccuracies();
		
		visitedWindows = new FPolygon2D[num_accuracies];
	}
	
	public void addStatus(ArcImsVectStatus status) throws ArcImsException {
		//Set an accuracy for the status
    	Rectangle2D geoExtent = status.getExtent();
		double iX = geoExtent.getWidth()/status.getWidth();
		double iY = geoExtent.getHeight()/status.getHeight();
		double accuracy = 1.0 * ((iX + iY)/2.0);
		
		int interval = -1;
		do{
			interval++;
		}
		while (accuracies[interval]<accuracy);
		
		
		//If visitedWindows element is null, create, add the status window and get the features
		if (visitedWindows[interval]==null) {
			GeneralPathX gp = new GeneralPathX((Shape) geoExtent);
			visitedWindows[interval] = new FPolygon2D(gp);
			features[interval] = getFeatures(status);
			}
		else{
			//If 
			
			
		}
		
		
		//If not, download the status not visited
		
		//Return geometries intersecting status
		
	}

	private HashMap getFeatures(ArcImsVectStatus status) throws ArcImsException {
		HashMap features = new HashMap();

		try {
			/*
			 * Build the request
			 */
			URL url = new URL(buildCapabilitiesRequest(status));
			/*
			 * Build the proper ArcXML
			 */
			String request = ArcXMLFeatures.getFeatureLayerRequest(status);
			if (status.verbose)
				System.err.println(request);
			/*
			 * Request the features
			 */
			logger.info("Start features retrieving");
			File response = ArcImsDownloadUtils.doRequestPost(url, request,
					"getFeatures.xml");
			/*
			 * Parse response and return array
			 */
			String layerId = (String) status.getLayerIds().get(0);
			logger.info("Start features parsing");
			features = getFeatures(new FileReader(response), status
					.getServiceInfo(), layerId, status.getSubfields(), true);
		} catch (MalformedURLException e) {
			logger.error(e.getMessage(), e);
			throw new ArcImsException("arcims_server_error");
		} catch (FileNotFoundException e) {
			logger.error(e.getMessage(), e);
			throw new ArcImsException("arcims_server_error");
		}
		return features;

	}

	private String buildCapabilitiesRequest(ArcImsStatus status)
    {
		StringBuffer req = new StringBuffer();
		
		
		String onlineResource;
		String service;
		onlineResource = status.getOnlineResource();
		service = status.getService();
		
		//String symbol = getSymbol(onlineResource);
		
		req.append(onlineResource);
		req.append("?"+GetVariables.SERVICENAME+"=");

		req.append(service);
		return req.toString();
    }

	public ArrayList getGeometries(ArcImsVectStatus status) {
		return null;
	}

	/**
	 * @return Returns the accuracies.
	 */
	public double[] getAccuracies() {
		return accuracies;
	}

	/**
	 * @param accuracies The accuracies to set.
	 */
	public void setAccuracies(double[] accuracies) {
		this.accuracies = accuracies;
	}
	
	/**
	 * Creates the double array of accuracies based on the
	 * fullextent width, number of pyramids to use and minimum
	 * status width stablished;
	 */
	private void createAccuracies(){
		//The minimum accuracy
		double min_acc = fullextent.getWidth()/MIN_STATUS_WIDTH;
		
		//The accuracy interval
		double interval = min_acc / num_accuracies;
		
		for (int i=0;i<num_accuracies;i++){
			accuracies[i] = interval * i;
		}
	}
	
	/**
	 * Method to add externally a new accuracy
	 * @param accuracy
	 */
	public void addAccuracy(double accuracy){
		num_accuracies++;
		FPolygon2D[] tempVisitedWindows = new FPolygon2D[num_accuracies];
		double[] tempAccuracies = new double[num_accuracies];
		
		int i = 0;
		do{
			tempAccuracies[i] = accuracies[i];
			tempVisitedWindows[i] = visitedWindows[i];
			i++;
		}
		while (accuracy>accuracies[i]);

		tempAccuracies[i]=accuracy;

		for (int j=i;j<num_accuracies;j++) {
			tempAccuracies[j+1]=accuracies[j];
			tempVisitedWindows[j+1]=visitedWindows[j];
		}
			
		
		setAccuracies(tempAccuracies);
		visitedWindows = tempVisitedWindows;
	}
	/**
	 * Parse the XML data retrieved from ArcIMS GET_FEATURES request
	 * to get a set of Features if withGeometries is true or an ArrayList of
	 * Value[] if this boolean is false
	 * @param f File to parse
	 * @param si ServiceInformation with metadata
	 * @param layerId Layer retrieved
	 * @param subfields Subfields retrieved
	 * @param withGeometries Will geometries be parsed?
	 * @return ArrayList of IFeatures or Value[]
	 * @throws ArcImsException
	 */
	private HashMap getFeatures(Reader lector, ServiceInformation si, String layerId, String[] subfields, boolean withGeometries) throws ArcImsException {
		FieldInformation fi = null;
		ServiceInformationLayerFeatures sil = 
			(ServiceInformationLayerFeatures) si.getLayerById(layerId);
		
		
		/*
		 * Prepare a Value array to store every FEATURE tag information
		 */
		Vector fieldsInfo = null;
		fieldsInfo = sil.getFieldsInfo();
		
		//If flag == true, throw an Exception (no layerId found)
		if (fieldsInfo==null)
			throw new ArcImsException("arcims_no_features");
		

		//We will store names and types in a hashmap
		ArrayList fieldsInfoA = new ArrayList();
		
		//Get a string array of names of fields to pass to the next level parser
//		for (int i=0;i<fieldsInfo.size();i++){
//			name = ((FieldInformation) fieldsInfo.get(i)).getName();
//			fi = ((FieldInformation) fieldsInfo.get(i));
//			fieldsInfoH.put(name,fi);
//		}
		
		/*
		 * If subfields[0] is equal to #ALL we add the entire vector
		 * into the Hasmap
		 */
			for (int i = 0; i<subfields.length;i++){
				fi = null;
				fi = sil.getFieldInformation(subfields[i]);
				if(fi!=null){
					fieldsInfoA.add(fi);
				}
			}
		
		
    	/*
    	 * Get the separators
    	 */
		String delCoords = si.getSeparators().getCs();
    	String delTuple = si.getSeparators().getTs();
    	char delDec = si.getSeparators().getDs();
    	
		/*
		 * ArrayList with features
		 */
		HashMap features = new HashMap();


		/*
		 * Start parsing
		 */
		int tag;
		KXmlParser kxmlParser = null;
		kxmlParser = new KXmlParser();
		//FileReader fr = null;

		
		IGeometry geom;
		Value[] values = new Value[fieldsInfoA.size()];
		int[] position = new int[fieldsInfoA.size()];
		
		//Initialize position array
		for (int i=0;i<position.length;i++)
			position[i] = -1;
		
		try {
//			fr = new FileReader(f);
			kxmlParser.setInput(lector);
			kxmlParser.nextTag();
			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				kxmlParser.require(KXmlParser.START_TAG, null,
						ServiceInfoTags.tARCXML);
				tag = kxmlParser.nextTag();
				while (tag != KXmlParser.END_DOCUMENT) {
					switch (tag) {
					case KXmlParser.START_TAG:
						if (kxmlParser.getName().compareTo(
							GetFeaturesTags.FEATURE) == 0) {
							/*
							 * parse FEATURES' tag
							 */
							//FIRST ID TAG
							values = parseValuesFromFeatureTag(
									kxmlParser, //the kxml parser
									fieldsInfoA,//Hashmap with names and FieldInformation objects)
									delDec, //the separators
									position //the relative positions of returned values
									);
							//If geometries are neede we add to the ArrayList the Ifeature
    						//NEXT GEOMETRIES
							geom = parseGeomFromFeatureTag(
									kxmlParser, //the kxml parser
									delTuple,delCoords,delDec //the separators
									);
							features.put(values,geom);		
								
						}
						 else if (kxmlParser.getName().compareTo(ServiceInfoTags.tERROR) == 0) {
							 logger.error("Error parsing GET_FEATURES:\r\n"+kxmlParser.nextText());
							throw new ArcImsException("arcims_server_error");
						}
						break;
					case KXmlParser.END_TAG:
						break;
					case KXmlParser.TEXT:

						break;
					}
					tag = kxmlParser.next();
				}

				kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
			}
		} catch (ParseException pe){
			logger.error(pe.getMessage(),pe);
			throw new ArcImsException("arcims_no_features");
		} catch (XmlPullParserException parser_ex) {
			logger.error(parser_ex.getMessage(),parser_ex);
			throw new ArcImsException("arcims_no_features");
		} catch (IOException ioe) {
			logger.error(ioe.getMessage(),ioe);
			throw new ArcImsException("arcims_no_features");
		}

		return features;
	}


	/**
	 * @param kxmlParser
	 * @param fieldsInfoH
	 * @param delDec
	 * @param position 
	 * @return
	 * @throws IOException 
	 * @throws XmlPullParserException 
	 * @throws ArcImsException 
	 * @throws ParseException 
	 */
	private Value[] parseValuesFromFeatureTag(KXmlParser parser,
			ArrayList fieldsInfoA, char delDec, int[] position)
			throws XmlPullParserException, IOException, ArcImsException,
			ParseException {
		int currentTag;
		boolean end = false;
		int f = 0;
		String att;
		String val;
		int type;

		Value[] values = new Value[fieldsInfoA.size()];

//		parser.require(KXmlParser.START_TAG,null,GetFeaturesTags.FIELDS);
		currentTag = parser.next();
		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().equals(GetFeaturesTags.FIELD)) {
					att = parser.getAttributeValue("", GetFeaturesTags.NAME);
					val = parser.getAttributeValue("", GetFeaturesTags.VALUE);

					/*
					 * As NAME is known, we need to get the position of that
					 * attribute into the ArrayList of fieldInformation objects
					 */
					if (position[f] == -1) {
						position[f] = getPosition(fieldsInfoA.iterator(), att);
					}
					if (position[f] == -1) {
						logger.error("Attribute not found at Metadata");
						throw new ArcImsException(
								"Attribute not found at Metadata");
					}

					// This way we can create a NullValue
					if (val.equals(""))
						val = null;

					/*
					 * At this point we know wat FieldInfo of the ArrayList we
					 * have to retrieve
					 */
					type = ((FieldInformation) fieldsInfoA.get(position[f]))
							.getType();

					// Add the Value into the correct position
					values[position[f]] = ArcImsValueFactory.createValueByType(
							val, type, delDec);

					// logger.info(att + " ("+type+")\t" +
					// values[f].toString());
					f++;
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().equals(GetFeaturesTags.FIELDS))
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return values;
	}



	/**
	 * Method to get the position of an attribute into the subfields array
	 * 
	 * @param fieldsInfoA
	 * @param att
	 * @return
	 */
	private int getPosition(Iterator it,String att) {
		int res = 0;
		FieldInformation fi;
		while (it.hasNext()){
			fi = (FieldInformation) it.next();
			if (fi.getName().equals(att))
				return res;
			res++;
		}
		//If no integer has returned at this point we return a -1
		return -1;
	}

	/**
	 * Parses the Feature Tag to get a IFeature
	 * @param KxmlParser with the FEATURE tag
	 * @param Value array with the correct subtypes to store FIELDS data
	 * @param String array with the correct fields names to retrieve
	 * @param String with the tuple separator
	 * @param String with the coordinates separator
	 * @param Char with the decimal separator
	 * @return @see com.hardcode.gdbms.engine.values.Value.IFeature
	 * @throws IOException
	 * @throws XmlPullParserException
	 * @throws ArcImsException
	 */
	private IGeometry parseGeomFromFeatureTag(KXmlParser parser, 
			String delTuple, String delCoords, char delDec)
			throws IOException, XmlPullParserException, ArcImsException {

		int currentTag;
		boolean end = false;


		IGeometry geom = null;

		currentTag = parser.next();
		while (!end) {
			
			String featType = parser.getName();
			
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (featType.equals(GetFeaturesTags.MULTIPOINT)) {
					// Go to COORDS tag
					parser.nextTag();

					if (parser.getName().equals(GetFeaturesTags.COORDS)) {
						// Create a tokenizer with the tuples
						String strPoints = parser.nextText();
						
						// Get the points from COORDS string
						ArrayList points = parseCoords(strPoints,delTuple,delCoords,delDec);
						
						//Convert these collections to array of doubles (primitive type)
						double[] xD = new double[points.size()];
						double[] yD = new double[points.size()];
						for (int i=0;i<xD.length;i++){
							xD[0]= ((Point2D) points.get(i)).getX();
							yD[0]= ((Point2D) points.get(i)).getY();
						}
						
						//Get the geometry
						geom = ShapeFactory.createMultipoint2D(xD,yD);
					} else {
						logger.error("Error parsing MULTIPOINT tag");
						throw new ArcImsException("arcims_features_error");
					}

				} 
				else if (featType.equals(GetFeaturesTags.POLYLINE) || featType.equals(GetFeaturesTags.POLYGON)){
					//Parse while parser doesn't found </FEATURES> tag
					//The GeneralPath to store the different paths
					GeneralPathX polyline = new GeneralPathX(GeneralPathX.WIND_EVEN_ODD);
					boolean endpol = false;
					while(!endpol){
						switch (currentTag){
						case KXmlParser.START_TAG:
							if (parser.getName().equals(GetFeaturesTags.COORDS)) {
								// Create a tokenizer with the tuples
								String strPoints = parser.nextText();
								
								// Get the points (Point2D) from COORDS string
								ArrayList points = parseCoords(strPoints,delTuple,delCoords,delDec);
								
								Iterator it = points.iterator();
								Point2D point;
								
								//First we MOVE to the first point of the path
								point = (Point2D) it.next();
								polyline.moveTo(point.getX(),point.getY());
								//And now we can LINE to the rest of the points of the path
								while (it.hasNext()){
									point = (Point2D) it.next();
									polyline.lineTo(point.getX(),point.getY());
								}
							}
							break;
						case KXmlParser.END_TAG:
							if (parser.getName().equals(featType))
								endpol=true;
							break;
						case KXmlParser.TEXT:
							break;
						}				
						if (!endpol)
							currentTag = parser.next();
					}
					if (featType.equals(GetFeaturesTags.POLYLINE)){
						geom = ShapeFactory.createPolyline2D(polyline);
					}
					else if (featType.equals(GetFeaturesTags.POLYGON)){
						geom = ShapeFactory.createPolygon2D(polyline);
					}
				}
				break;
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(GetFeaturesTags.FEATURE) == 0)
					end = true;
				break;
			case KXmlParser.TEXT:
				break;
			}
			if (!end)
				currentTag = parser.next();
		}
		return geom;
	}


	/**
	 * Private method that parses a COORDS tag to return an ArrayList of Point2D.Double objects
	 * @param strCoords
	 * @param delTuple
	 * @param delCoords
	 * @param delDec
	 * @return
	 */
	private ArrayList parseCoords(String strCoords, String delTuple, String delCoords, char delDec) {
		String tupla = new String();
		ArrayList points = new ArrayList();
	    String[] tuplaPart;
		Double x;
		Double y;
		
		/*
		 * Replaces Decimal Separator to convert coords to suitable doubles.
		 * 
		 * Maybe it will have a better performance if we insert this replace into
		 * the while
		 */
		if (delDec!='.') strCoords = strCoords.replace(delDec,'.');
		
		
		//Creates a tokenizer to run over the string
		StringTokenizer tuplas = new StringTokenizer(strCoords,delTuple,false);

		//Get ArrayLists collections with X's and Y's
		while (tuplas.hasMoreTokens()){
			tupla = tuplas.nextToken();
			tuplaPart = tupla.split(delCoords);
			x = new Double(tuplaPart[0]);
			y = new Double(tuplaPart[1]);
			
			//Version with in-loop replacing
			/*  
			if (delDec!='.'){
			x = new Double(tuplaPart[0].replace(delDec,'.'));
			y = new Double(tuplaPart[1].replace(delDec,'.'));
			} else{
			x = new Double(tuplaPart[0]);
			y = new Double(tuplaPart[1]);
		    } 
			 */
			
			points.add(new Point2D.Double(x.doubleValue(), y.doubleValue()));
		}	
		//Return an array op Point2D objects
		return points;
	}
}
